GroundProximityDetection = true

config = {
	["ah_64"] = {},
	["mh_60"] = {},
	["mi_17"] = {},
	["mi_24"] = {weaponLoc = Vec(-0.05, 0.95, -9), music = "MOD/katyusha.ogg"},
	["uh_1"] = {weaponLoc = Vec(0.0, -0.9, -3.3), music = "MOD/fortunate_son.ogg"},
	["ka_50"] = {weaponLoc = Vec(0.85, 1.1, -3.9), music = "MOD/soviet_anthem.ogg", rotors = "coaxial"},
	["av_8"] = {weaponLoc = Vec(-0.55, 0.65, -2.5)},
	["mh_60medic"] = {},
	["delorean"] = {}
}

function EulerQuat(q, k)
	local x, y, z, w = q[1], q[2], q[3], q[4]
	local bank, heading, attitude

	local s = 2 * x * y + 2 * z * w
	if s >= 1 then
		bank = 0
		heading = 2 * math.atan2(x, w)
		attitude = math.pi / 2
	elseif s <= -1 then
		bank = 0
		heading = -2 * math.atan2(x, w)
		attitude = math.pi / -2
	else
		bank = math.atan2(2 * x * w - 2 * y * z, 1 - 2 * x ^ 2 - 2 * z ^ 2)
		heading = math.atan2(2 * y * w - 2 * x * z, 1 - 2 * y ^ 2 - 2 * z ^ 2)
		attitude = math.asin(s)
	end
	bank, heading, attitude = math.deg(bank), math.deg(heading), math.deg(attitude)

	if k and k > 90 and k < 270 then
		bank, heading, attitude = bank - 180, heading - 180, 180 - attitude
	elseif k and k >= 270 then
		attitude = attitude + 360
	elseif k and k < -90 then
		bank, heading, attitude = bank - 180, heading - 180, -(attitude + 180)
	end
	return bank, heading, attitude
end

function clamp(value, mi, ma)
	if value < mi then value = mi end
	if value > ma then value = ma end
	return value
end

function sign(x)
	if x < 0 then
		return -1
	elseif x > 0 then
		return 1
	else
		return 0
	end
end

function YawRight(yaw)
	yaw = yaw - 1.0
	if yaw < -180 then
		yaw = yaw + 360
	end
	return yaw
end

function YawLeft(yaw)
	yaw = yaw + 1.0
	if yaw > 180 then
		yaw = yaw - 360
	end
	return yaw
end

function init()
	chopper = {}
	chopper.body = FindBody("chopper")
	chopper.transform = GetBodyTransform(chopper.body)

	chopper.mainRotor = FindBody("mainrotor")
	chopper.mainRotorLocalTransform = TransformToLocalTransform(chopper.transform, GetBodyTransform(chopper.mainRotor))

	chopper.tailRotor = FindBody("tailrotor")
	chopper.tailRotorLocalTransform = TransformToLocalTransform(chopper.transform, GetBodyTransform(chopper.tailRotor))

	chopper.angle = 0.0
	chopper.mass = GetBodyMass(chopper.body)
	chopper.sound = LoadLoop("chopper-loop.ogg")

	shootSound = LoadSound("chopper-shoot0.ogg")
	rocketSound = LoadSound("tools/launcher0.ogg")

	pitch = 0
	yaw = 0
	roll = 0
	collective = chopper.transform.pos[2] + 0.5 -- Starting altitude

	cameraX = 0
	cameraY = 0
	zoom = 20

	vehicleType = GetTagValue(chopper.body, "chopper")
	if config[vehicleType].music then
		music = config[vehicleType].music
	end
	if not HasTag(chopper.body, "avf") and config[vehicleType].weaponLoc then
		weaponLoc = config[vehicleType].weaponLoc
	end
	if HasTag(chopper.body, "avf") then
		isAVF = true
	end
	isCoaxial = false
	if config[vehicleType].rotors then
		isCoaxial = true
	end

	playing = true -- Default: false
	aimMode = false
	hover = false
	dynamicRotor = true
end


function tick(dt)
	local vehicle = GetPlayerVehicle()
	if vehicle ~= FindVehicle("helicopter") then
		if not dynamicRotor then -- Player leaves the chopper
			dynamicRotor = true
			SetBodyDynamic(chopper.mainRotor, true)
			SetBodyAngularVelocity(chopper.body, Vec())
			SetPlayerVehicle(0)
			pitch = 0
			roll = 0
			cameraY = pitch
			local playerSpawn = TransformToParentPoint(chopper.transform, Vec(2, 0, -4))
			local playerRot = QuatEuler(pitch, yaw, roll)
			SetPlayerTransform(Transform(playerSpawn, playerRot))
		end
		return
	elseif GetBodyMass(chopper.body) < chopper.mass * 0.66 then
		if not dynamicRotor then -- Helicopter is too damaged
			dynamicRotor = true
			SetBodyDynamic(chopper.mainRotor, true)
		end
		return
	else
		if dynamicRotor then -- Player enter the chopper
			dynamicRotor = false
			SetBodyDynamic(chopper.mainRotor, false)
			chopper.transform.pos = GetBodyTransform(chopper.body).pos -- Update chopper position
			local unused = 0
			unused, yaw = GetQuatEuler(GetBodyTransform(chopper.body).rot) -- Update yaw rotation
			cameraX = yaw
			collective = chopper.transform.pos[2] + 0.5
		end
		if music and not playing then
			playing = true
			PlayMusic(music)
		end
	end

	if InputDown("w") then
		collective = collective + 0.1
	end
	if InputDown("s") and collective > 0 then
		collective = collective - 0.1
	end

	if InputDown("a") and roll < 30 then
		roll = roll + 0.5
	end
	if InputDown("d") and roll > -30 then
		roll = roll - 0.5
	end

	if weaponLoc then
		local firePos = TransformToParentPoint(chopper.transform, weaponLoc)
		local fireDir = TransformToParentVec(chopper.transform, Vec(0, 0, -1))
		if vehicleType == "mh_60" then
			fireDir = VecNormalize(TransformToParentVec(chopper.transform, Vec(2, -1, 0)))
			if InputDown("lmb") then
				PlaySound(shootSound, chopper.transform.pos, 5)
				Shoot(firePos, fireDir, 0)
			end
		else
			if InputPressed("lmb") then
				PlaySound(shootSound, chopper.transform.pos, 5)
				Shoot(firePos, fireDir, 0)
			end
			if InputPressed("rmb") then
				PlaySound(rocketSound, chopper.transform.pos, 5)
				Shoot(firePos, fireDir, 1)
			end
		end
	elseif isAVF then
		if InputPressed("rmb") then
			aimMode = not aimMode
		end
	end
	if InputPressed("h") then
		hover = not hover
	end

	local prevYaw = yaw

	if not aimMode then
		local x, y = GetQuatEuler(GetCameraTransform().rot)

		if x < pitch and pitch > -45 then
			pitch = pitch - 0.5
		end
		if x > pitch and pitch < 15 then
			pitch = pitch + 0.5
		end

		if y < yaw and not (y < yaw - 180) then
			yaw = YawRight(yaw)
		elseif y < yaw - 180 then
			yaw = YawLeft(yaw)
		end

		if y > yaw and not (y > yaw + 180) then
			yaw = YawLeft(yaw)
		elseif y > yaw + 180 then
			yaw = YawRight(yaw)
		end
	else
		local mouseX = -InputValue("mousedx")
		local mouseY = -InputValue("mousedy")

		if sign(mouseY) < 0 and pitch > -45 then
			pitch = pitch - 0.5
		elseif sign(mouseY) > 0 and pitch < 15 then
			pitch = pitch + 0.5
		end

		if sign(mouseX) < 0 then
			yaw = YawRight(yaw)
		elseif sign(mouseX) > 0 then
			yaw = YawLeft(yaw)
		end
	end

	if hover then
		if pitch ~= 0 then
			pitch = pitch - sign(pitch)
		end
		yaw = prevYaw
	end

	local moveDir = Vec(0.01 * -roll, 0, 0.01 * pitch)
	local pos = TransformToParentPoint(chopper.transform, moveDir)

	if GroundProximityDetection then
		pos[2] = pos[2] - 0.1
		local dir = VecNormalize(VecSub(pos, chopper.transform.pos))
		local maxDist = 0.5 + math.abs(pitch / 5) + math.abs(roll / 10)
		QueryRejectBody(chopper.body)
		QueryRequire("physical large")
		local hit, dist = QueryRaycast(chopper.transform.pos, dir, maxDist)
		--local posEnd = VecAdd(chopper.transform.pos, VecScale(dir, maxDist))
		if hit then
			collective = collective + (maxDist - dist) / (2 * maxDist)
		end
	end
	pos[2] = collective

	chopper.angle = chopper.angle + 0.6
	local rot = QuatEuler(math.sin(chopper.angle*0.053)*5 + pitch, math.sin(chopper.angle*0.04)*5 + yaw, roll)

	if aimMode then
		rot = QuatEuler(pitch, yaw, roll)
	end

	chopper.transform.pos = pos
	chopper.transform.rot = rot
	--SetBodyTransform(chopper.body, chopper.transform)
	

	chopper.mainRotorLocalTransform.rot = QuatEuler(0, chopper.angle * 57, 0)
	--SetBodyTransform(chopper.mainRotor, TransformToParentTransform(chopper.transform, chopper.mainRotorLocalTransform))

	if isCoaxial then
		chopper.tailRotorLocalTransform.rot = QuatEuler(0, -chopper.angle * 57, 0)
	else
		chopper.tailRotorLocalTransform.rot = QuatEuler(chopper.angle * 57, 0, 0)
	end
	--SetBodyTransform(chopper.tailRotor, TransformToParentTransform(chopper.transform, chopper.tailRotorLocalTransform))

	PlayLoop(chopper.sound, chopper.transform.pos, 10)

	local mx, my = InputValue("mousedx"), InputValue("mousedy")
	cameraX = cameraX - mx / 10
	cameraY = cameraY - my / 10
	cameraY = clamp(cameraY, -90, 90)
	local cameraRot = QuatEuler(cameraY, cameraX, 0)
	local cameraT = Transform(chopper.transform.pos, cameraRot)
	zoom = zoom - InputValue("mousewheel")
	zoom = clamp(zoom, 10, 30)
	local cameraPos = TransformToParentPoint(cameraT, Vec(0, 5, zoom))
	local camera = Transform(cameraPos, cameraRot)
	SetCameraTransform(camera)
end

function update(dt)
	local vehicle = GetPlayerVehicle()
	if vehicle == FindVehicle("helicopter") and not dynamicRotor then
		SetBodyTransform(chopper.body, chopper.transform)
		SetBodyTransform(chopper.mainRotor, TransformToParentTransform(chopper.transform, chopper.mainRotorLocalTransform))
		SetBodyTransform(chopper.tailRotor, TransformToParentTransform(chopper.transform, chopper.tailRotorLocalTransform))
	end
end
